/*
 * This file is part of the Anthony Lomax Java Library.
 *
 * Copyright (C) 2003 Anthony Lomax <anthony@alomax.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
The TauP Toolkit: Flexible Seismic Travel-Time and Raypath Utilities.
Copyright (C) 1998-2000 University of South Carolina

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

The current version can be found at
<A HREF="www.seis.sc.edu">http://www.seis.sc.edu</A>

Bug reports and comments should be directed to
H. Philip Crotwell, crotwell@seis.sc.edu or
Tom Owens, owens@seis.sc.edu

 */
/*
 * TauP_Table_NLL.java
 *
 * Created on 01 December 2003, 11:08
 */
package edu.sc.seis.TauP;

import java.io.*;
import java.util.*;

/**
 *
 * @author  Anthony Lomax
 */
public class TauP_Table_NLL extends TauP_Table {

    public static final int NONLINLOC = 2;
    public static final int NONLINLOC_ANGLES = 3;
    protected String nllGridDesc = "181,0.0,180.0,61,0.0,600.0";
    protected int numDistance;
    protected double distance0;
    protected double deltaDistance;
    protected int numDepth;
    protected double depth0;
    protected double deltaDepth;
    protected String stationName = "DEFAULT";
    protected double stationLong = 0.0;
    protected double stationLat = 0.0;
    protected double stationDepth = 0.0;

    /** Creates a new instance of TauP_Table_NLL */
    public TauP_Table_NLL() throws TauModelException {
        super();
    }

    public void init() throws IOException {
        super.init();
    }

    public void start() throws TauModelException, TauPException, IOException {
        switch (outputType) {
            case TauP_Table.GENERIC:
                genericTable(dos);
                break;
            case TauP_Table.LOCSAT:
                locsatTable(dos);
                break;
            case TauP_Table_NLL.NONLINLOC:
                dos.close();	// not needed for NLL
                nonLinLocTable();
                break;
            case TauP_Table_NLL.NONLINLOC_ANGLES:
                dos.close();	// not needed for NLL
                nonLinLocTable();
                break;
            default:
                throw new TauPException(
                        "TauP_Table: undefined state for output type: " + outputType);
        }
    }

    protected void nonLinLocTable() throws TauPException, TauModelException, IOException {

        parseGridHeader();

        DataOutputStream bufOut;
        DataOutputStream slownessBufOut = null;

        int numDistanceSlowness = 3;

        char materialProperty = (((PhaseName) phaseNames.elementAt(0)).getName()).charAt(0);

        String timeFileRoot = outFile + "." + phaseNames.elementAt(0) + "." + stationName;
        String type;
        if (outputType == TauP_Table_NLL.NONLINLOC_ANGLES) {
            timeFileRoot += ".angle";
            type = "ANGLE2D";
        } else {
            timeFileRoot += ".time";
            type = "TIME2D";
        }

        // open output file
        if (outFile != null && outFile.length() != 0) {
            bufOut = new DataOutputStream(new BufferedOutputStream(
                    new FileOutputStream(timeFileRoot + ".buf")));
            PrintStream hdrOut = new PrintStream(new FileOutputStream(timeFileRoot + ".hdr"));
            writeGridHeader(hdrOut, numDistance, type);
            hdrOut.close();
            // slowness model file (used by NLLoc LOCMETH OT_STACK, etc.)
            if ((outputType == TauP_Table_NLL.NONLINLOC) && (materialProperty == 'p' || materialProperty == 'P')) {
                String slownessFileRoot = outFile + ".P.mod";
                slownessBufOut = new DataOutputStream(new BufferedOutputStream(
                        new FileOutputStream(slownessFileRoot + ".buf")));
                hdrOut = new PrintStream(new FileOutputStream(slownessFileRoot + ".hdr"));
                writeGridHeader(hdrOut, numDistanceSlowness, "SLOW_LEN");
                hdrOut.close();
            } else if ((outputType == TauP_Table_NLL.NONLINLOC) && (materialProperty == 's' || materialProperty == 'S')) {
                String slownessFileRoot = outFile + ".S.mod";
                slownessBufOut = new DataOutputStream(new BufferedOutputStream(
                        new FileOutputStream(slownessFileRoot + ".buf")));
                hdrOut = new PrintStream(new FileOutputStream(slownessFileRoot + ".hdr"));
                writeGridHeader(hdrOut, numDistanceSlowness, "SLOWNESS");
                hdrOut.close();
            }

        } else {
            bufOut = new DataOutputStream(System.out);
            writeGridHeader(System.out, numDistance, type);
        }

        float[][] timeArray = null;
        short[][][] angleArray = null;
        if (outputType == TauP_Table_NLL.NONLINLOC_ANGLES) {
            angleArray = new short[numDepth][numDistance][2];
        } else {
            timeArray = new float[numDepth][numDistance];
        }
        float[][] slownessArray = null;
        if (slownessBufOut != null) {
            slownessArray = new float[numDepth][numDistanceSlowness];
        }

        // loop over depth (z)
        for (int depthNum = 0; depthNum < numDepth; depthNum++) {
            double depth = depth0 + deltaDepth * (double) depthNum;
            depthCorrect(depth0 + deltaDepth * (double) depthNum);
            double velocity = 0.0;
            try {
                int layerNumber = tMod.sMod.vMod.layerNumberBelow(depth);
                VelocityLayer velocityLayer = tMod.sMod.vMod.getVelocityLayer(layerNumber);
                velocity = velocityLayer.evaluateAt(depth, materialProperty);
            } catch (Exception e) {
                throw (new TauPException(e.getMessage()));
            }
            // loop over distance (x)
            for (int distNum = 0; distNum < numDistance; distNum++) {
                //
                if (slownessBufOut != null && distNum < numDistanceSlowness) {
                    slownessArray[depthNum][distNum] = (float) (1.0 / velocity);
                }
                //
                calculate(distance0 + deltaDistance * (double) distNum);
                Arrival[] arrivals = getArrivals();
                double time = Double.MAX_VALUE;
                Arrival minArrival = null;
                for (int aNum = 0; aNum < getNumArrivals(); aNum++) {
                    Arrival currArrival = arrivals[aNum];
                    if (currArrival.time < time) {
                        time = currArrival.time;
                        minArrival = currArrival;
                    }
                }
                if (outputType == TauP_Table_NLL.NONLINLOC_ANGLES) {
                    /* 1983__Buland_Chapman__The_Computation_of_Seismic_Travel_Times__BSSA
                     *  rayParam = r * sin(i) / velocity
                     *  thus,
                     *  i = asin(rayParam * velocity / r)
                     */
                    double dip = (180.0 / Math.PI) * Math.asin(minArrival.rayParam * velocity / tMod.sMod.getRadiusOfEarth());
                    // check down- or up-going
                    boolean downgoing = true;
                    int branch = minArrival.phase.tMod.getSourceBranch();
                    while (true) {
                        try {
                            downgoing = ((Boolean) (minArrival.phase.downGoing.elementAt(branch))).booleanValue();
                            break;
                        } catch (Exception e) {
                            if (branch > 0) {
                                branch--;
                                continue;
                            }
                            System.out.println("downgoing error: " + e);
                            break;
                        }
                    }
                    if (!downgoing) {
                        dip = 180.0 - dip;
                    }
                    setTakeOffAngles(angleArray[depthNum][distNum], 1.0, dip, 9);
                    //System.out.println("dist " + (distance0 + deltaDistance * (double) distNum) + " d " + depth + " v " + velocity
                    //+ " rp " + minArrival.rayParam + " dip " + dip + " downgoing " + downgoing);
                } else {
                    if (time == Double.MAX_VALUE) {
                        time = -1.0;
                    }
                    timeArray[depthNum][distNum] = (float) time;
                }
            }
        }

        // write to grid file
        if (outputType == TauP_Table_NLL.NONLINLOC_ANGLES) {
            // loop over distance (y)
            for (int distNum = 0; distNum < numDistance; distNum++) {
                // loop over depth (z)
                for (int depthNum = 0; depthNum < numDepth; depthNum++) {
                    bufOut.writeShort(angleArray[depthNum][distNum][0]);
                    bufOut.writeShort(angleArray[depthNum][distNum][1]);
                    //System.out.println("i0 " + angleArray[depthNum][distNum][0] + " i1 " + angleArray[depthNum][distNum][1]);
                }
            }
        } else {
            // loop over distance (y)
            for (int distNum = 0; distNum < numDistance; distNum++) {
                // loop over depth (z)
                for (int depthNum = 0; depthNum < numDepth; depthNum++) {
                    bufOut.writeFloat(timeArray[depthNum][distNum]);
                }
            }
            if (slownessBufOut != null) {
                // loop over distance (y)
                for (int distNum = 0; distNum < numDistanceSlowness; distNum++) {
                    // loop over depth (z)
                    for (int depthNum = 0; depthNum < numDepth; depthNum++) {
                        slownessBufOut.writeFloat(slownessArray[depthNum][distNum]);
                    }
                }
            }
        }

        bufOut.close();
        if (slownessBufOut != null) {
            slownessBufOut.close();
        }
    }

    /*** function to set angle values in take-off angles union */
    public void setTakeOffAngles(short[] ival, double azim, double dip, int iqual) {

        int iazim = (int) Math.round(10.0 * azim);
        int idip = (int) Math.round(10.0 * dip);

        // remove sign bits
        iazim = (iazim << 1) >>> 1;
        idip = (idip << 1) >>> 1;
        iqual = (iqual << 1) >>> 1;

        ival[0] = (short) iazim;
        ival[1] = (short) (idip * 16 + iqual);

        //System.out.println("iazim " + iazim + " idip " + idip + " iqual " + iqual + " i0 " + ival[0] + " i1 " + ival[1]);

    }

    /** writes NLL Grid Header file contents to hdrOUT */
    public void writeGridHeader(PrintStream hdrOut, int numDistanceHdr, String type) {

        // write NLL grid header lines
        hdrOut.print(1);
        hdrOut.print(" ");
        hdrOut.print(numDistanceHdr);
        hdrOut.print(" ");
        hdrOut.print(numDepth);
        hdrOut.print("  ");

        hdrOut.print(0.0);
        hdrOut.print(" ");
        hdrOut.print(distance0);
        hdrOut.print(" ");
        hdrOut.print(depth0);
        hdrOut.print("  ");

        hdrOut.print(0.0);
        hdrOut.print(" ");
        hdrOut.print(deltaDistance);
        hdrOut.print(" ");
        hdrOut.print(deltaDepth);
        hdrOut.print("  ");

        hdrOut.print(type);
        hdrOut.print("  ");

        hdrOut.print("FLOAT");

        hdrOut.println();

        hdrOut.println(stationName + " " + stationLong + " " + stationLat + " " + stationDepth);

    }

    /** parse NLL Grid Header string */
    public void parseGridHeader() throws TauPException {

        double distance1;
        double depth1;

        // parse grid description
        StringTokenizer stkzr = new StringTokenizer(nllGridDesc, ",");
        try {
            // "181,0.0,180.0,61,0.0,600.0"
            numDistance = Integer.parseInt(stkzr.nextToken());
            distance0 = Double.parseDouble(stkzr.nextToken());
            distance1 = Double.parseDouble(stkzr.nextToken());
            numDepth = Integer.parseInt(stkzr.nextToken());
            depth0 = Double.parseDouble(stkzr.nextToken());
            depth1 = Double.parseDouble(stkzr.nextToken());
        } catch (Exception e) {
            throw new TauPException(
                    "TauP_Table_NLL: error parsing grid description string: " + nllGridDesc);
        }
        deltaDistance = (distance1 - distance0) / (double) (numDistance - 1);
        deltaDepth = (depth1 - depth0) / (double) (numDepth - 1);
    }

    public void printUsage() {
        printStdUsageHead();
        System.out.println(
                "-ph phase list     -- comma separated phase list\n"
                + "-pf phasefile      -- file containing phases\n\n"
                + "-mod[el] modelname -- use velocity model \"modelname\" for calculations\n"
                + "                      Default is iasp91.\n\n");

        System.out.println(
                "-header filename   -- reads depth and distance spacing data\n"
                + "                      from a LOCSAT style file.");
        System.out.println(
                "-generic           -- outputs a \"generic\" ascii table\n");
        System.out.println(
                "-locsat            -- outputs a \"locsat\" style ascii table\n");
        System.out.println(
                "-nll griddesc      -- outputs a \"NonLinLoc\" 3D Grid Data buffer file for each phase\n");
        printStdUsageTail();
    }

    public String[] parseCmdLineArgs(String[] args) throws IOException {
        int i = 0;
        String[] leftOverArgs;
        int numNoComprendoArgs = 0;
        File tempFile;

        leftOverArgs = super.parseCmdLineArgs(args);
        String[] noComprendoArgs = new String[leftOverArgs.length];

        while (i < leftOverArgs.length) {
            if (leftOverArgs[i].equals("-nll")) {
                outputType = NONLINLOC;
                nllGridDesc = args[i + 1];
                i++;
            } else if (leftOverArgs[i].equals("-nll_angles")) {
                outputType = NONLINLOC_ANGLES;
                nllGridDesc = args[i + 1];
                i++;
            } else {
                noComprendoArgs[numNoComprendoArgs++] = leftOverArgs[i];
            }
            i++;
        }

        if (numNoComprendoArgs > 0) {
            String[] temp = new String[numNoComprendoArgs];
            System.arraycopy(noComprendoArgs, 0, temp, 0, numNoComprendoArgs);
            return temp;
        } else {
            return new String[0];
        }
    }

    public static void main(String[] args) {
        TauP_Table_NLL me;
        try {
            me = new TauP_Table_NLL();
            String[] noComprendoArgs = me.parseCmdLineArgs(args);
            if (noComprendoArgs.length > 0) {
                for (int i = 0; i < noComprendoArgs.length; i++) {
                    if (noComprendoArgs[i].equals("-help")
                            || noComprendoArgs[i].equals("-version")) {
                        System.exit(0);
                    }
                }
                System.out.println("I don't understand the following arguments, continuing:");
                for (int i = 0; i < noComprendoArgs.length; i++) {
                    System.out.print(noComprendoArgs[i] + " ");
                    if (noComprendoArgs[i].equals("-help")
                            || noComprendoArgs[i].equals("-version")) {
                        System.out.println();
                        System.exit(0);
                    }
                }
                System.out.println();
                noComprendoArgs = null;
            }

            me.init();
            me.start();
        } catch (Exception e) {
            System.err.println("Caught Exception: " + e);
            e.printStackTrace();
        }
    }
}
